import argparse
import logging
from rich.logging import RichHandler
import yaml

from gaarf.io import writer, reader  # type: ignore
from gaarf.simulation import simulate_data, SimulatorSpecification
from .utils import GaarfConfigBuilder, ConfigSaver, initialize_runtime_parameters


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("-s",
                        "--simulator-config",
                        dest="simulator_config",
                        default=None)
    parser.add_argument("--account", dest="customer_id", default="None")
    parser.add_argument("--output", dest="save", default="console")
    parser.add_argument("--input", dest="input", default="file")
    parser.add_argument("--api-version", dest="api_version", default=12)
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--customer-ids-query",
                        dest="customer_ids_query",
                        default=None)
    parser.add_argument("--customer-ids-query-file",
                        dest="customer_ids_query_file",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.set_defaults(save_config=False)
    args = parser.parse_known_args()
    main_args = args[0]

    logging.basicConfig(format="%(message)s",
                        level=main_args.loglevel.upper(),
                        datefmt="%Y-%m-%d %H:%M:%S",
                        handlers=[RichHandler(rich_tracebacks=True)])
    logging.getLogger("google.ads.googleads.client").setLevel(logging.WARNING)
    logging.getLogger("smart_open.smart_open_lib").setLevel(logging.WARNING)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    logger = logging.getLogger(__name__)

    config = GaarfConfigBuilder(args).build()
    logger.debug("config: %s", config)

    writer_client = writer.WriterFactory().create_writer(
        config.output, **config.writer_params)
    if config.output == "bq":
        _ = writer_client.create_or_get_dataset()
    reader_factory = reader.ReaderFactory()
    reader_client = reader_factory.create_reader(main_args.input)

    for query in main_args.query:
        if simulator_config := main_args.simulator_config:
            with open(simulator_config, encoding="utf-8") as f:
                simulator_config = yaml.safe_load(f)
            simulator_specification = SimulatorSpecification(
                **simulator_config)
        else:
            simulator_specification = SimulatorSpecification()
        report = simulate_data(reader_client.read(query),
                               simulator_specification)
        writer_client.write(report, query)


if __name__ == "__main__":
    main()



import pytest

from gaarf.simulation import simulate_data, SimulatorSpecification
from gaarf.query_editor import QuerySpecification


@pytest.fixture
def query():
    return """
        SELECT
            segments.date AS date,
            campaign.advertising_channel_sub_type AS campaign_type,
            campaign.id as campaign_id,
            campaign.name AS campaign_name,
            asset.youtube_video_asset.youtube_video_id AS video_id
        FROM campaign
        """


@pytest.fixture
def query_specification(query):
    return QuerySpecification(query).generate()


@pytest.fixture
def default_simulator_specification():
    return SimulatorSpecification()


@pytest.fixture
def default_report(query, default_simulator_specification):
    return simulate_data(query, default_simulator_specification)


def test_n_rows_have_been_generated(default_report,
                                    default_simulator_specification):
    assert len(default_report) == default_simulator_specification.n_rows


def test_n_cols_have_been_generated(default_report, query_specification):
    assert len(default_report.results[0]) == len(query_specification.fields)


@pytest.fixture
def allowed_enums():
    return ["APP_CAMPAIGN", "APP_CAMPAIGN_FOR_ENGAGEMENT"]


@pytest.fixture
def replacements():
    return {"asset.youtube_video_asset.youtube_video_id": ["12345", "54321"]}


@pytest.fixture
def custom_simulator_specification(allowed_enums, replacements):
    return SimulatorSpecification(
        allowed_enums={"campaign.advertising_channel_sub_type": allowed_enums},
        replacements=replacements)


@pytest.fixture
def custom_report(query, custom_simulator_specification):
    return simulate_data(query, custom_simulator_specification)


def test_allowed_enums_were_simulated(custom_report,
                                      custom_simulator_specification,
                                      allowed_enums):
    report_iterator = iter(custom_report)
    assert next(report_iterator).get("campaign_type") in allowed_enums
    assert next(report_iterator).get("campaign_type") in allowed_enums


def test_replacements_were_applied(custom_report, custom_simulator_specification,
                      replacements):
    report_iterator = iter(custom_report)
    replacement = replacements.get("asset.youtube_video_asset.youtube_video_id")
    assert next(report_iterator).get("video_id") in replacement
    assert next(report_iterator).get("video_id") in replacement



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Set, Sequence
from dataclasses import dataclass

from google.ads.googleads.client import GoogleAdsClient  # type: ignore
from importlib import import_module
from pathlib import Path
import re
from proto.primitives import ProtoType


@dataclass
class Field:
    name: str
    type: str
    values: Set[str]


class BaseClient:

    def __init__(self, version: str = "v12"):
        self.api_version = version
        self.google_ads_row = self._get_google_ads_row(version)

    def get_response(self, entity_id: str, query_text: str):
        pass

    def _get_google_ads_row(self, api_version: str) -> "GoogleAdsRow":
        base_module = f"google.ads.googleads.{api_version}"
        google_ads_service = import_module(
            f"{base_module}.services.types.google_ads_service")
        return google_ads_service.GoogleAdsRow()

    def infer_types(self, fields: Sequence[str]) -> List[Any]:
        """Maps API fields to Python primitives."""

        base_module = f"google.ads.googleads.{self.api_version}"
        common_types_module = f"{base_module}.common.types"
        segments = import_module(f"{common_types_module}.segments")
        metrics = import_module(f"{common_types_module}.metrics")

        mapping = {"INT64": int, "FLOAT": float, "DOUBLE": float, "BOOL": bool}
        output = []
        for field in fields:
            try:
                resource, *sub_resource, base_field = field.split(".")
                base_field = "type_" if base_field == "type" else base_field
                values = set()
                if resource == "metrics":
                    target_resource = metrics.Metrics
                elif resource == "segments":
                    # If segment has name segments.something.something
                    if sub_resource:
                        target_resource = getattr(
                            segments, f"{clean_resource(sub_resource[-1])}")
                    else:
                        target_resource = getattr(segments,
                                                  f"{clean_resource(resource)}")
                else:
                    resource_module = import_module(
                        f"{base_module}.resources.types.{resource}")

                    target_resource = getattr(resource_module,
                                              f"{clean_resource(resource)}")
                    try:
                        # If resource has name resource.something.something
                        if sub_resource:
                            target_resource = getattr(
                                target_resource,
                                f"{clean_resource(sub_resource[-1])}")
                    except AttributeError:
                        resource_module = import_module(
                            f"{base_module}.resources.types.{sub_resource[0]}")
                        if len(sub_resource) > 1:
                            if hasattr(resource_module,
                                       f"{clean_resource(sub_resource[1])}"):
                                target_resource = getattr(
                                    resource_module,
                                    f"{clean_resource(sub_resource[-1])}")
                            else:
                                resource_module = import_module(
                                    f"{common_types_module}.ad_type_infos")

                                target_resource = getattr(
                                    resource_module,
                                    f"{clean_resource(sub_resource[1])}Info")
                        else:
                            target_resource = getattr(
                                resource_module,
                                f"{clean_resource(sub_resource[-1])}")
                descriptor = target_resource.meta.fields.get(
                    base_field).descriptor
                result = descriptor.type

                if result == 14:  # 14 stands for ENUM
                    enum_class, enum = descriptor.type_name.split(".")[-2:]
                    file_name = re.sub(r'(?<!^)(?=[A-Z])', '_', enum).lower()
                    enum_resource = import_module(
                        f"{base_module}.enums.types.{file_name}")
                    values = set([
                        p.name
                        for p in getattr(getattr(enum_resource, enum_class), enum)
                    ])

                field_type = mapping.get(ProtoType(result).name, str)
                if result != 14:
                    if field_type == str:
                        values = {
                            "",
                        }
                    if field_type == int:
                        values = {
                            0,
                        }
                    if field_type == float:
                        values = {
                            0.0,
                        }
                    if field_type == bool:
                        values = {
                            False,
                        }
                field = Field(name=field, type=field_type, values=values)
            except (AttributeError, ModuleNotFoundError):
                field = Field(name=field, type=str, values={
                    "",
                })
            output.append(field)
        return output


class GoogleAdsApiClient(BaseClient):
    default_google_ads_yaml = str(Path.home() / "google-ads.yaml")

    def __init__(self,
                 path_to_config: str = default_google_ads_yaml,
                 config_dict: Dict[str, Any] = None,
                 yaml_str: str = None,
                 version: str = "v12"):
        super().__init__(version)
        self.client = self._init_client(path=path_to_config,
                                        config_dict=config_dict,
                                        yaml_str=yaml_str,
                                        version=version)
        if hasattr(self.client, "use_proto_plus"):
            self.client.use_proto_plus = True
        else:
            raise ValueError("Specify 'use_proto_plus: True' in your google-ads.yaml file")
        self.ads_service = self.client.get_service("GoogleAdsService")

    def get_response(self, entity_id, query_text):
        response = self.ads_service.search_stream(customer_id=entity_id,
                                                  query=query_text)
        return response

    def _init_client(self, path, config_dict, yaml_str,
                     version) -> Optional[GoogleAdsClient]:
        if config_dict:
            return GoogleAdsClient.load_from_dict(config_dict, version)
        if yaml_str:
            return GoogleAdsClient.load_from_string(yaml_str, version)
        if path:
            return GoogleAdsClient.load_from_storage(path, version)
        try:
            return GoogleAdsClient.load_from_env(version)
        except Exception as e:
            raise ValueError("Cannot instantiate GoogleAdsClient")


def clean_resource(resource: str) -> str:
    return resource.title().replace('_', '')



import pytest
from gaarf.bq_executor import extract_datasets, expand_jinja


def test_extract_datasets():
    macros = {
        "start_date": ":YYYYMMDD",
        "bq_dataset": "dataset_1",
        "dataset_new": "dataset_2",
        "legacy_dataset_old": "dataset_3",
        "wrong_dts": "dataset_4"
    }

    expected = ["dataset_1", "dataset_2", "dataset_3"]
    datasets = extract_datasets(macros)
    assert datasets == expected


@pytest.fixture
def templated_query():
    return "SELECT field_one, {% for day in cohort_days %} {{day}} AS day_{{day}}, {% endfor %} FROM some_table"


def test_expand_jinja_if_template_parameters_are_provided(templated_query):
    expected_query = "SELECT field_one, 1 AS day_1, 2 AS day_2, FROM some_table"
    rendered_query= expand_jinja(templated_query, cohort_days="1,2")
    assert rendered_query.replace("  ", " ") == expected_query


def test_expand_jinja_if_empty_template_parameter_is_provided(templated_query):
    expected_query = "SELECT field_one, FROM some_table"
    rendered_query= expand_jinja(templated_query, cohort_days=None)
    assert rendered_query.replace("  ", " ") == expected_query


def test_expand_jinja_not_template_parameter_is_provided(templated_query):
    expected_query = "SELECT field_one, FROM some_table"
    rendered_query = expand_jinja(templated_query)
    assert rendered_query.replace("  ", " ") == expected_query



import pytest
import pandas as pd
from collections import abc
from gaarf.report import GaarfReport, GaarfRow


@pytest.fixture
def single_element_report():
    return GaarfReport(results=[1], column_names=["campaign_id"])


@pytest.fixture
def single_column_report():
    return GaarfReport(results=[[1], [2], [3]], column_names=["campaign_id"])


@pytest.fixture
def multi_column_report():
    return GaarfReport(results=[[1, 2], [2, 3], [3, 4]],
                       column_names=["campaign_id", "ad_group_id"])


def test_single_element_report_returns_sequence(single_element_report):
    assert [row for row in single_element_report] == [1]


def test_single_column_report_returns_sequence(single_column_report):
    assert [row for row in single_column_report] == [1, 2, 3]


def test_multi_column_report_returns_gaarf_row(multi_column_report):
    results = [row for row in multi_column_report]
    assert isinstance(results[0], GaarfRow)


def test_multi_column_report_support_iteration_with_gaarf_iterator(
        multi_column_report):
    assert isinstance(iter(multi_column_report), abc.Iterable)


def test_multi_column_report_get_element_by_id(multi_column_report):
    assert [row[0] for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_by_name(multi_column_report):
    assert [row["campaign_id"] for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_with_get_method(multi_column_report):
    assert [row.get("campaign_id") for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_as_attribute(multi_column_report):
    assert [row.campaign_id for row in multi_column_report] == [1, 2, 3]


def test_getitem_raise_index_error_for_out_of_index_value(multi_column_report):
    with pytest.raises(IndexError):
        [row[99] for row in multi_column_report] == [1, 2, 3]


def test_get_raises_attribute_error_for_missing_value(multi_column_report):
    with pytest.raises(AttributeError):
        [row.get("missing_value")
         for row in multi_column_report] == [None, None, None]


def test_getattr_raises_attribute_error_for_missing_value(multi_column_report):
    with pytest.raises(AttributeError):
        [getattr(row, "missing_value")
         for row in multi_column_report] == [None, None, None]


def test_hasattr_return_false_for_missing_value(multi_column_report):
    assert [hasattr(row, "missing_value")
            for row in multi_column_report] == [False, False, False]


def test_convert_report_to_pandas(multi_column_report):
    expected = pd.DataFrame(data=[[1, 2], [2, 3], [3, 4]],
                            columns=["campaign_id", "ad_group_id"])
    assert multi_column_report.to_pandas().equals(expected)


def test_get_report_length(multi_column_report):
    assert len(multi_column_report) == 3


def test_report_str(single_element_report):
    assert str(single_element_report) == "[1]"


def test_report_bool(single_element_report):
    assert single_element_report
    single_element_report.is_fake = True
    assert not single_element_report


def test_add_two_reports(multi_column_report):
    added_report = multi_column_report + multi_column_report
    assert len(added_report) == 6


def test_add_report_and_non_report_raises_type_error(multi_column_report):
    with pytest.raises(TypeError):
        multi_column_report + 1


def test_add_non_report_and_report_raises_type_error(multi_column_report):
    with pytest.raises(TypeError):
        1 + multi_column_report


def test_add_reports_with_different_columns_raises_value_error(
        multi_column_report, single_element_report):
    with pytest.raises(ValueError):
        multi_column_report + single_element_report


def test_iteration_over_fake_report_returns_empty_list(multi_column_report):
    multi_column_report.is_fake = True
    results = [row for row in multi_column_report]
    assert results == []


def test_slicing_multi_column_gaarf_report_returns_gaarf_report(
        multi_column_report):
    new_report = multi_column_report[0:2]
    assert new_report == GaarfReport(
        results=[[1, 2], [2, 3]], column_names=["campaign_id", "ad_group_id"])


def test_indexing_multi_column_gaarf_report_by_single_index_returns_gaarf_row(
        multi_column_report):
    new_report = multi_column_report[0]
    assert new_report == GaarfRow(data=[1, 2],
                                  column_names=["campaign_id", "ad_group_id"])


def test_indexing_multi_column_gaarf_report_by_multi_index_returns_gaarf_report(
        multi_column_report):
    new_report = multi_column_report[0:2]
    assert new_report == GaarfReport(
        results=[[1, 2], [2, 3]], column_names=["campaign_id", "ad_group_id"])


def test_indexing_multi_column_gaarf_report_by_one_column_returns_gaarf_report(
        multi_column_report):
    new_report = multi_column_report["campaign_id"]
    assert new_report == GaarfReport(results=[[1], [2], [3]],
                                     column_names=["campaign_id"])


def test_indexing_multi_column_gaarf_report_by_several_columns_returns_gaarf_report(
        multi_column_report):
    new_report = multi_column_report[["campaign_id", "ad_group_id"]]
    assert new_report == multi_column_report


def test_indexing_multi_column_gaarf_report_by_non_existing_column_raises_type_error(
        multi_column_report):
    with pytest.raises(TypeError):
        multi_column_report[["campaign_id", "ad_group"]]


def test_slicing_multi_column_gaarf_report_returns_slice(single_column_report):
    new_report = single_column_report[0:2]
    assert new_report == [1, 2]


def test_slicing_multi_column_gaarf_report_returns_element(
        single_column_report):
    new_report = single_column_report[0]
    assert new_report == [1]



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Sequence, Union
from collections import abc
import operator
import pandas as pd


class GaarfReport:

    def __init__(self,
                 results: Sequence[Any],
                 column_names: Sequence[str],
                 is_fake: bool = False):
        self.results = results
        self.column_names = column_names
        self.multi_column_report = len(column_names) > 1
        self.is_fake = is_fake

    def to_list(self) -> Sequence[Any]:
        return self.results

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(data=self.results, columns=self.column_names)

    def __len__(self):
        return len(self.results)

    def __iter__(self):
        if self.is_fake:
            return None
        for result in self.results:
            if self.multi_column_report:
                yield GaarfRow(result, self.column_names)
            elif isinstance(result, abc.Sequence):
                yield result[0]
            else:
                yield result

    def __bool__(self):
        return not self.is_fake

    def __str__(self):
        return f"{self.results}"

    def __getitem__(self, key):
        cls = type(self)
        if isinstance(key, abc.MutableSequence):
            if set(key).issubset(set(self.column_names)):
                indices = []
                for k in key:
                    indices.append(self.column_names.index(k))
                results = []
                for row in self.results:
                    rows = []
                    for index in indices:
                        rows.append(row[index])
                    results.append(rows)
                return cls(results, key)
            else:
                non_existing_keys = set(key).intersection(set(self.column_names))
                if len(non_existing_keys) > 1:
                    message =  f"Columns '{', '.join(list(non_existing_keys))}' cannot be found in the report"
                message =  f"Column '{non_existing_keys.pop()}' cannot be found in the report"
                raise TypeError(message)
        else:
            if key in self.column_names:
                index = self.column_names.index(key)
                results = [[row[index]] for row in self.results]
                return cls(results, [key])
        if self.multi_column_report:
            if isinstance(key, slice):
                return cls(self.results[key], self.column_names)
            return GaarfRow(self.results[key], self.column_names)
        if isinstance(key, slice):
            return [element[0] for element in self.results[key]]
        index = operator.index(key)
        return self.results[key]

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.column_names != other.column_names:
            return False
        return self.results == other.results

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Add operation is supported only for GaarfReport")
        if self.column_names != other.column_names:
            raise ValueError("column_names should be the same in GaarfReport")
        return GaarfReport(results=self.results + other.results,
                           column_names=self.column_names)


class GaarfRow:

    def __init__(self, data: Sequence[Union[int, float, str]],
                 column_names: Sequence[str]):
        self.data = data
        try:
            self.n_elements = len(data)
        except TypeError:
            self.n_elements = 1
        self.column_names = column_names

    def __getattr__(self, element: str) -> Any:
        if element in self.column_names:
            return self.data[self.column_names.index(element)]
        raise AttributeError(f"cannot find {element} element!")

    def __getitem__(self, element: Union[str, int]) -> Any:
        if isinstance(element, int) and element < self.n_elements:
            return self.data[element]
        if isinstance(element, str):
            return self.__getattr__(element)
        raise IndexError(f"cannot find {element} element!")

    def get(self, item: str) -> Any:
        return self.__getattr__(item)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.column_names != other.column_names:
            return False
        return self.data == other.data

    def __repr__(self):
        dict_data = {x[1]: x[0] for x in zip(self.data, self.column_names)}
        return f"GaarfRow(\n{dict_data}\n)"



import pytest
from typing import Any, Dict, List
from dataclasses import dataclass
from proto import Message
import gaarf.parsers as parsers
from gaarf.query_editor import VirtualColumn, VirtualColumnError


@dataclass
class FakeGoogleAdsRowElement:
    value: int
    text: str
    name: str


@dataclass
class TextAttribute:
    text: str


@dataclass
class NameAttribute:
    name: str


@dataclass
class ValueAttribute:
    value: str


@dataclass
class Metric:
    clicks: int
    impressions: int


@dataclass
class FakeAdsRowMultipleElements:
    campaign_type: NameAttribute
    clicks: int
    resource: str
    value: str
    metrics: Metric


@dataclass
class FakeQuerySpecification:
    customizers: Dict[str, Any]
    virtual_columns: Dict[str, Any]
    fields: List[str]
    column_names: List[str]


@pytest.fixture
def fake_query_specification():
    customizers = {
        "resource": {
            "type": "resource_index",
            "value": 0
        },
        "value": {
            "type": "nested_field",
            "value": "value"
        }
    }
    virtual_columns = {"date": {"type": "built-in", "value": "date"}}
    return FakeQuerySpecification(
        customizers=customizers,
        virtual_columns=virtual_columns,
        fields=["campaign_type", "clicks", "resource", "value"],
        column_names=["campaign_type", "clicks", "resource", "value"])


@pytest.fixture
def fake_google_ads_row_element():
    return FakeGoogleAdsRowElement(1, "2", "3")


@pytest.fixture
def google_ads_row_parser(fake_query_specification):
    return parsers.GoogleAdsRowParser(fake_query_specification)


def test_google_ads_row_parser_return_last_parser_in_chain(
        google_ads_row_parser):
    assert type(google_ads_row_parser.parser) == parsers.RepeatedParser


def test_google_ads_row_parser_chooses_correct_element_parser(
        google_ads_row_parser, fake_google_ads_row_element):
    assert google_ads_row_parser.parse(fake_google_ads_row_element) == "3"


def test_resource_formatter_get_resource():
    resource = parsers.ResourceFormatter.get_resource("name: id")
    assert resource == "id"


def test_resource_formatter_get_resource_id():
    resource = parsers.ResourceFormatter.get_resource_id(
        "customers/1/resource/2")
    assert resource == "2"


def test_resource_formatter_clear_resource_id_int():
    resource = parsers.ResourceFormatter.clean_resource_id('"1"')
    assert resource == 1


def test_resource_formatter_clear_resource_id_str():
    resource = parsers.ResourceFormatter.clean_resource_id('"value"')
    assert resource == "value"


@pytest.fixture
def base_parser():
    return parsers.BaseParser(None)


def test_base_parser(base_parser):
    assert base_parser.parse("") == None


@pytest.fixture
def text_attribute():
    return TextAttribute("some-text")


@pytest.fixture
def name_attribute():
    return NameAttribute("some-name")


@pytest.fixture
def value_attribute():
    return ValueAttribute(1)


def test_attribute_parser(base_parser, text_attribute, name_attribute,
                          value_attribute):
    attribute_parser = parsers.AttributeParser(base_parser)
    assert attribute_parser.parse("") == None
    assert attribute_parser.parse(name_attribute) == "some-name"
    assert attribute_parser.parse(text_attribute) == "some-text"
    assert attribute_parser.parse(value_attribute) == 1


@pytest.mark.skip("WIP")
def test_empty_attribute_parser(base_parser):
    empty_attribute_parser = parsers.EmptyAttributeParser(base_parser)
    assert empty_attribute_parser.parse("") == None
    message = Message()
    assert empty_attribute_parser.parse(message) == "Not set"


def test_google_ads_row_parser(google_ads_row_parser, text_attribute,
                               name_attribute, value_attribute):
    assert isinstance(google_ads_row_parser.parser, parsers.RepeatedParser)
    assert google_ads_row_parser.parse(name_attribute) == "some-name"
    assert google_ads_row_parser.parse(text_attribute) == "some-text"
    assert google_ads_row_parser.parse(value_attribute) == 1
    assert google_ads_row_parser.parse("some-value") == None


@pytest.fixture
def fake_ads_row():
    return FakeAdsRowMultipleElements(campaign_type=NameAttribute("SEARCH"),
                                      clicks=1,
                                      resource="customers/1/resource/2",
                                      value=ValueAttribute(1),
                                      metrics=Metric(clicks=10,
                                                     impressions=10))


def test_get_attributes_from_row(google_ads_row_parser, fake_ads_row,
                                 fake_query_specification):
    extracted_rows = google_ads_row_parser._get_attributes_from_row(
        fake_ads_row, google_ads_row_parser.row_getter)
    assert extracted_rows == (NameAttribute("SEARCH"), 1,
                              "customers/1/resource/2", ValueAttribute(1))
    assert google_ads_row_parser.parse_ads_row(fake_ads_row) == [
        "SEARCH", 1, "2", 1
    ]


def test_convert_builtin_virtual_column(google_ads_row_parser,
                                           fake_ads_row):
    fake_builtin_virtual_column = VirtualColumn(type="built-in",
                                                      value="fake_value")
    result = google_ads_row_parser._convert_virtual_column(
        fake_ads_row, fake_builtin_virtual_column)
    assert result == "fake_value"


@pytest.fixture
def fake_expression_virtual_column():
    return VirtualColumn(
        type="expression",
        value="metrics.clicks / metrics.impressions",
        fields=["metrics.clicks", "metrics.impressions"],
        substitute_expression="{metrics_clicks} / {metrics_impressions}")


def test_convert_expression_virtual_column(
        google_ads_row_parser, fake_ads_row,
        fake_expression_virtual_column):
    result = google_ads_row_parser._convert_virtual_column(
        fake_ads_row, fake_expression_virtual_column)
    assert result == 1.0


def test_convert_expression_virtual_column_with_zero_denominator_returns_zero(
        google_ads_row_parser, fake_ads_row, fake_expression_virtual_column):
    fake_ads_row.metrics.impressions = 0
    result = google_ads_row_parser._convert_virtual_column(
        fake_ads_row, fake_expression_virtual_column)
    assert result == 0


def test_if_convert_expression_virtual_column_raises_type_error_raise_virtual_column_error(
        google_ads_row_parser, fake_ads_row, fake_expression_virtual_column):
    fake_ads_row.metrics.impressions = "str"
    with pytest.raises(VirtualColumnError):
        result = google_ads_row_parser._convert_virtual_column(
            fake_ads_row, fake_expression_virtual_column)


def test_if_convert_expression_virtual_column_fails_return_column_value(
        google_ads_row_parser, fake_ads_row, fake_expression_virtual_column):
    fake_ads_row.metrics.impressions = "0 +"  # this should raise SyntaxError
    result = google_ads_row_parser._convert_virtual_column(
        fake_ads_row, fake_expression_virtual_column)
    assert result == fake_expression_virtual_column.value




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

from collections.abc import MutableSequence
from concurrent import futures
import argparse
from pathlib import Path
import logging
from rich.logging import RichHandler
from smart_open import open
import yaml

from gaarf import api_clients, utils, query_executor
from gaarf.io import writer, reader  # type: ignore
from .utils import GaarfConfigBuilder, ConfigSaver, initialize_runtime_parameters, gaarf_runner


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="*")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--account", dest="customer_id", default=None)
    parser.add_argument("--output", dest="save", default="console")
    parser.add_argument("--input", dest="input", default="file")
    parser.add_argument("--ads-config",
                        dest="config",
                        default=str(Path.home() / "google-ads.yaml"))
    parser.add_argument("--api-version", dest="api_version", default=12)
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--customer-ids-query",
                        dest="customer_ids_query",
                        default=None)
    parser.add_argument("--customer-ids-query-file",
                        dest="customer_ids_query_file",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.add_argument("--parallel-queries",
                        dest="parallel_queries",
                        action="store_true")
    parser.add_argument("--no-parallel-queries",
                        dest="parallel_queries",
                        action="store_false")
    parser.add_argument("--optimize-performance",
                        dest="optimize_performance",
                        default="NONE")
    parser.add_argument("--dry-run", dest="dry_run", action="store_true")
    parser.add_argument("--disable-account-expansion",
                        dest="disable_account_expansion",
                        action="store_true")
    parser.add_argument("-v", "--version", dest="version", action="store_true")
    parser.set_defaults(save_config=False)
    parser.set_defaults(parallel_queries=True)
    parser.set_defaults(dry_run=False)
    parser.set_defaults(disable_account_expansion=False)
    args = parser.parse_known_args()
    main_args = args[0]

    if main_args.version:
        import pkg_resources
        version = pkg_resources.require(
            "google-ads-api-report-fetcher")[0].version
        print(f"gaarf version {version}")
        exit()

    logging.basicConfig(format="%(message)s",
                        level=main_args.loglevel.upper(),
                        datefmt="%Y-%m-%d %H:%M:%S",
                        handlers=[RichHandler(rich_tracebacks=True)])
    logging.getLogger("google.ads.googleads.client").setLevel(logging.WARNING)
    logging.getLogger("smart_open.smart_open_lib").setLevel(logging.WARNING)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    logger = logging.getLogger(__name__)

    if not main_args.query:
        logger.error("Please provide one or more queries to run")
        exit()

    with open(main_args.config, "r", encoding="utf-8") as f:
        google_ads_config_dict = yaml.safe_load(f)

    config = GaarfConfigBuilder(args).build()
    if account := main_args.customer_id:
        config.account = account
    if not config.account:
        if mcc := google_ads_config_dict.get("login_customer_id"):
            config.account = str(mcc)
        else:
            raise ValueError(
                "No account found, please specify via --account CLI flag"
                "or add as login_customer_id in google-ads.yaml")
    logger.debug("config: %s", config)

    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)
    if main_args.dry_run:
        exit()

    if config.params:
        config = initialize_runtime_parameters(config)
    logger.debug("initialized config: %s", config)

    ads_client = api_clients.GoogleAdsApiClient(
        config_dict=google_ads_config_dict, version=f"v{config.api_version}")
    reader_factory = reader.ReaderFactory()
    reader_client = reader_factory.create_reader(main_args.input)

    if config.customer_ids_query:
        customer_ids_query = config.customer_ids_query
    elif config.customer_ids_query_file:
        file_reader = reader_factory.create_reader("file")
        customer_ids_query = file_reader.read(config.customer_ids_query_file)
    else:
        customer_ids_query = None

    if main_args.disable_account_expansion:
        logger.info(
            "Skipping account expansion because of disable_account_expansion flag"
        )
        customer_ids = config.account if isinstance(
            config.account, MutableSequence) else [config.account]
    else:
        customer_ids = utils.get_customer_ids(ads_client, config.account,
                                              customer_ids_query)
    if customer_ids:
        writer_client = writer.WriterFactory().create_writer(
            config.output, **config.writer_params)
        if config.output == "bq":
            _ = writer_client.create_or_get_dataset()
        ads_query_executor = query_executor.AdsQueryExecutor(ads_client)

        logger.info("Total number of customer_ids is %d, accounts=[%s]",
                    len(customer_ids), ",".join(map(str, customer_ids)))

        if main_args.parallel_queries:
            logger.info("Running queries in parallel")
            with futures.ThreadPoolExecutor() as executor:
                future_to_query = {
                    executor.submit(ads_query_executor.execute,
                                    reader_client.read(query), query,
                                    customer_ids, writer_client, config.params,
                                    main_args.optimize_performance): query
                    for query in main_args.query
                }
                for future in futures.as_completed(future_to_query):
                    query = future_to_query[future]
                    gaarf_runner(query, future.result, logger)
        else:
            logger.info("Running queries sequentially")
            for query in main_args.query:
                callback = ads_query_executor.execute(
                    reader_client.read(query), query, customer_ids,
                    writer_client, config.params,
                    main_args._optimize_performance)
                gaarf_runner(query, callback, logger)
    else:
        logger.warning(
            "Not a single under MCC %s is found that satisfies the following customer_id query: '%s'",
            config.account, customer_ids_query)


if __name__ == "__main__":
    main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import proto

from typing import Sequence, Union
from collections.abc import MutableSequence

from .api_clients import GoogleAdsApiClient
from .query_editor import QuerySpecification
from .query_executor import AdsReportFetcher
from .report import GaarfRow


def get_customer_ids(ads_client: GoogleAdsApiClient,
                     customer_id: Union[str, MutableSequence[str]],
                     customer_ids_query: str = None) -> Sequence[str]:
    """Gets list of customer_ids from an MCC account.

    Args:
        ads_client: GoogleAdsApiClient used for connection.
        customer_id: MCC account_id.
        custom_query: GAQL query used to reduce the number of customer_ids.
    Returns:
        All customer_ids from MCC safisfying the condition.
    """

    query = """
    SELECT customer_client.id FROM customer_client
    WHERE customer_client.manager = FALSE AND customer_client.status = "ENABLED"
    """
    query_specification = QuerySpecification(query).generate()
    if not isinstance(customer_id, MutableSequence):
        customer_id = customer_id.split(",")
    report_fetcher = AdsReportFetcher(ads_client, customer_id)
    customer_ids = report_fetcher.fetch(query_specification).to_list()
    if customer_ids_query:
        report_fetcher = AdsReportFetcher(ads_client, customer_ids)
        query_specification = QuerySpecification(customer_ids_query).generate()
        customer_ids = report_fetcher.fetch(query_specification)
        customer_ids = [
            row[0] if isinstance(row, GaarfRow) else row
            for row in customer_ids
        ]

    customer_ids = list(
        set([customer_id for customer_id in customer_ids if customer_id != 0]))

    return customer_ids



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import abc
from typing import Any, Dict, Tuple, Sequence, Union
from operator import attrgetter
import datetime
import re
import proto  # type: ignore
from proto.marshal.collections.repeated import (  # type: ignore
    Repeated, RepeatedComposite)  # type: ignore
from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer,
    RepeatedCompositeFieldContainer)  # type: ignore

from .query_editor import VirtualColumn, VirtualColumnError


class BaseParser:

    def __init__(self, successor):
        self._successor = successor

    def parse(self, request):
        if self._successor:
            return self._successor.parse(request)
        return None


class RepeatedParser(BaseParser):

    def parse(self, request):
        if isinstance(
                request,
            (Repeated,
             RepeatedScalarFieldContainer)) and "customer" in str(request):
            elements = []
            for request_element in request:
                element = ResourceFormatter.get_resource_id(request_element)
                elements.append(ResourceFormatter.clean_resource_id(element))
            return elements
        return super().parse(request)


class RepeatedCompositeParser(BaseParser):

    def parse(self, request):
        if isinstance(request,
                      (RepeatedComposite, RepeatedCompositeFieldContainer)):
            elements = []
            for request_element in request:
                resource = ResourceFormatter.get_resource(request_element)
                element = ResourceFormatter.get_resource_id(resource)
                elements.append(ResourceFormatter.clean_resource_id(element))
            return elements
        return super().parse(request)


class AttributeParser(BaseParser):

    def parse(self, request):
        if hasattr(request, "name"):
            return request.name
        if hasattr(request, "text"):
            return request.text
        if hasattr(request, "value"):
            return request.value
        return super().parse(request)


class EmptyAttributeParser(BaseParser):

    def parse(self, request):
        if issubclass(type(request), proto.Message):
            return "Not set"
        return super().parse(request)


class GoogleAdsRowParser:

    def __init__(self, query_specification, nested_fields=None):
        self.nested_fields = nested_fields
        self.parser = self._init_parsers()
        self.row_getter = attrgetter(*query_specification.fields)
        self.fields = query_specification.fields
        self.customizers = query_specification.customizers
        self.virtual_columns = query_specification.virtual_columns
        self.column_names = query_specification.column_names

    def _init_parsers(self):
        parser_chain = BaseParser(None)
        for parser in [
                EmptyAttributeParser, AttributeParser, RepeatedCompositeParser,
                RepeatedParser
        ]:
            new_parser = parser(parser_chain)
            parser_chain = new_parser
        return parser_chain

    def parse(self, request):
        return self.parser.parse(request)

    def parse_ads_row(self, row) -> Sequence[Any]:
        final_rows = []
        extracted_attributes = self._get_attributes_from_row(
            row, self.row_getter)
        index = 0
        virtual_columns = {}
        for i, column in enumerate(self.column_names):
            if column in self.virtual_columns.keys():
                parsed_element = self._convert_virtual_column(
                    row, self.virtual_columns[column])
            else:
                extracted_attribute = extracted_attributes[index]
                index += 1
                if self.customizers:
                    if (caller := self.customizers.get(column)):
                        if caller.get("type") == "nested_field":
                            values_ = caller.get("value").split(".")
                            extracted_attribute_ = getattr(
                                extracted_attribute, values_[0]) if hasattr(
                                    extracted_attribute,
                                    values_[0]) else extracted_attribute
                            try:
                                if isinstance(
                                        extracted_attribute,
                                    (Repeated, RepeatedComposite,
                                     RepeatedScalarFieldContainer,
                                     RepeatedCompositeFieldContainer
                                     )) or isinstance(
                                         extracted_attribute_,
                                         (Repeated, RepeatedComposite,
                                          RepeatedScalarFieldContainer,
                                          RepeatedCompositeFieldContainer)):
                                    if isinstance(
                                            extracted_attribute_,
                                        (Repeated, RepeatedComposite)):
                                        extracted_attribute = extracted_attribute_
                                    if len(values_) > 1:
                                        value = values_[1]
                                    else:
                                        value = caller.get("value")
                                    extracted_attribute = list(
                                        set([
                                            attrgetter(value)(element)
                                            for element in extracted_attribute
                                        ]))
                                else:
                                    extracted_attribute = attrgetter(
                                        caller.get("value"))(
                                            extracted_attribute)
                            except AttributeError as e:
                                raise ValueError(
                                    f"customizer {caller} is incorrect,\ndetails: '{e}',\n"
                                    f"row: '{row}'")
                        elif caller.get("type") == "resource_index":
                            extracted_attribute = re.split(
                                "~", extracted_attribute)[caller.get("value")]
                            extracted_attribute = re.split(
                                "/", extracted_attribute)[-1]
                if isinstance(extracted_attribute, abc.MutableSequence):
                    parsed_element = [
                        self.parser.parse(element) or element
                        for element in extracted_attribute
                    ]
                else:
                    parsed_element = self.parser.parse(
                        extracted_attribute) or extracted_attribute
            final_rows.append(parsed_element)
        return final_rows if len(final_rows) > 1 else final_rows[0]

    def _get_attributes_from_row(self, row, getter) -> Tuple[Any, ...]:
        attributes = getter(row)
        return attributes if isinstance(attributes, tuple) else (attributes, )

    def _convert_virtual_column(self, row,
                                   virtual_column: VirtualColumn) -> str:
        if virtual_column.type == "built-in":
            return virtual_column.value
        elif virtual_column.type == "expression":
            virtual_column_getter = attrgetter(*virtual_column.fields)
            virtual_column_values = virtual_column_getter(row)
            try:
                iter(virtual_column_values)
            except TypeError:
                virtual_column_values = (virtual_column_values, )
            virtual_column_replacements = {
                field.replace(".", "_"): value
                for field, value in zip(virtual_column.fields,
                                        virtual_column_values)
            }
            try:
                result = eval(
                    virtual_column.substitute_expression.format(
                        **virtual_column_replacements))
            except ZeroDivisionError:
                return 0
            except TypeError:
                raise VirtualColumnError(
                    f"cannot parse virtual_column {virtual_column.value}"
                )
            except Exception as e:
                return virtual_column.value
            return result
        else:
            raise ValueError(
                f"Unsupported virtual column type: {virtual_column_type}"
            )


class ResourceFormatter:

    @staticmethod
    def get_resource(element: str) -> str:
        return re.split(": ", str(element).strip())[1]

    @staticmethod
    def get_resource_id(element: str) -> str:
        return re.split("/", str(element))[-1]

    @staticmethod
    def clean_resource_id(element: str) -> Union[int, str]:
        element = re.sub('"', '', str(element))
        try:
            return int(element)
        except ValueError:
            return element



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union
from google.cloud import bigquery  # type: ignore
from google.cloud.exceptions import NotFound  # type: ignore
from jinja2 import Template
import logging
import pandas as pd

logger = logging.getLogger(__name__)


class BigQueryExecutorException(Exception):
    pass


class BigQueryExecutor:

    def __init__(self, project_id: str, location: Optional[str] = None):
        self.project_id = project_id
        self.location = location
        self.client = bigquery.Client(project_id)

    def execute(
            self,
            script_name: str,
            query_text: str,
            params: Optional[Dict[str, Any]] = None) -> Optional[pd.DataFrame]:
        logger.debug("Original query text:\n%s", query_text)
        if params:
            if (templates := params.get("template")):
                query_templates = {
                    name: value
                    for name, value in templates.items() if name in query_text
                }
                if query_templates:
                    query_text = expand_jinja(query_text, **query_templates)
                    logger.debug("Query text after jinja expansion:\n%s",
                                 query_text)
            if (macros := params.get("macro")):
                query_text = query_text.format(**macros)
                logger.debug("Query text after macro substitution:\n%s",
                             query_text)
        job = self.client.query(query_text)
        try:
            result = job.result()
            logger.debug("%s launched successfully", script_name)
            if result.total_rows:
                return result.to_dataframe()
            return None
        except Exception as e:
            raise BigQueryExecutorException(e) from e

    def create_datasets(self, macros: Optional[Dict[str, Any]]) -> None:
        if macros:
            if (datasets := extract_datasets(macros)):
                for dataset in datasets:
                    dataset_id = f"{self.project_id}.{dataset}"
                    try:
                        bq_dataset = self.client.get_dataset(dataset_id)
                    except NotFound:
                        bq_dataset = bigquery.Dataset(dataset_id)
                        bq_dataset.location = self.location
                        bq_dataset = self.client.create_dataset(bq_dataset,
                                                                timeout=30)
                        logger.debug("Created new dataset %s", dataset_id)


def extract_datasets(macros: Optional[Dict[str, Any]]) -> Optional[List[str]]:
    if not macros:
        return None
    return [value for macro, value in macros.items() if "dataset" in macro]


def expand_jinja(query_text, **template_params):
    for key, value in template_params.items():
        if value:
            if len(splitted_param := value.split(",")) > 1:
                template_params[key] = splitted_param
            else:
                template_params[key] = [value]
        else:
            template_params = ""
    template = Template(query_text)
    return template.render(template_params)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from concurrent import futures
from enum import Enum, auto
import itertools
from google.ads.googleads.errors import GoogleAdsException  # type: ignore
import logging
from google.api_core import exceptions
from tenacity import Retrying, RetryError, retry_if_exception_type, stop_after_attempt, wait_exponential

from . import parsers
from . import api_clients
from .query_editor import QuerySpecification, QueryElements
from .report import GaarfReport
from .io import writer, reader  # type: ignore

logger = logging.getLogger(__name__)


class OptimizeStrategy(Enum):
    NONE = auto()
    BATCH = auto()
    PROTOBUF = auto()
    BATCH_PROTOBUF = auto()


class AdsReportFetcher:
    """Class responsible for getting data from Ads API.

    Attributes:
        api_client: a client used for connecting to Ads API.
        customer_ids: account_id(s) from which data will be fetches.
    """

    def __init__(self, api_client: api_clients.BaseClient,
                 customer_ids: Union[Sequence[str], str]) -> None:

        self.api_client = api_client
        self.customer_ids = [
            customer_ids
        ] if not isinstance(customer_ids, list) else customer_ids

    def fetch(self,
              query_specification: Union[str, QueryElements],
              optimize_strategy: str = "NONE") -> GaarfReport:
        """Fetches data from Ads API based on query_specification.

        Args:
            query_specification: query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            GaarfReport with results of query execution.
        """

        total_results: List[List[Tuple[Any]]] = []
        is_fake_report = False
        if not isinstance(query_specification, QueryElements):
            query_specification = QuerySpecification(
                text=str(query_specification),
                api_version=self.api_client.api_version).generate()

        parser = parsers.GoogleAdsRowParser(query_specification)
        for customer_id in self.customer_ids:
            logger.debug("Running query %s for customer_id %s",
                         query_specification.query_title, customer_id)
            try:
                results = self._parse_ads_response(
                    query_specification, customer_id, parser,
                    getattr(OptimizeStrategy, optimize_strategy))
                total_results.extend(results)
                if query_specification.is_constant_resource:
                    logger.debug("Constant resource query: running only once")
                    break
            except GoogleAdsException as e:
                logger.error("Cannot execute query %s for %s",
                             query_specification.query_title, customer_id)
                logger.error(str(e))
        if not total_results:
            row = self.api_client.google_ads_row
            results = [parser.parse_ads_row(row)]
            total_results.extend(results)
            is_fake_report = True
            logger.warning(
                "Query %s generated zero results, using placeholders to infer schema",
                query_specification.query_title)
        return GaarfReport(results=total_results,
                           column_names=query_specification.column_names,
                           is_fake=is_fake_report)

    def _parse_ads_response(
        self,
        query_specification: QueryElements,
        customer_id: str,
        parser: parsers.GoogleAdsRowParser,
        optimize_strategy: OptimizeStrategy = OptimizeStrategy.NONE
    ) -> List[List[Tuple[Any]]]:
        """Parses response returned from Ads API request.

        Args:
            query_specification: query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_id: account for which data should be requested.
            parser: instance of parser class that transforms each row from
                request into desired format.
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            Parsed rows for the whole response.
        """
        total_results: List[List[Tuple[Any]]] = []
        logger.debug("Getting response for query %s for customer_id %s",
                     query_specification.query_title, customer_id)
        try:
            for attempt in Retrying(retry=retry_if_exception_type(
                    exceptions.InternalServerError),
                                    stop=stop_after_attempt(3),
                                    wait=wait_exponential()):
                with attempt:
                    response = self.api_client.get_response(
                        entity_id=str(customer_id),
                        query_text=query_specification.query_text)
        except RetryError as retry_failure:
            logger.error("Cannot fetch data from API for query '%s' %d times",
                         query_specification.query_title,
                         retry_failure.last_attempt.attempt_number)
        if optimize_strategy in (OptimizeStrategy.BATCH,
                                 OptimizeStrategy.BATCH_PROTOBUF):
            logger.warning("Running gaarf in an optimized mode")
            logger.warning("Optimize strategy is %s", optimize_strategy.name)
            if optimize_strategy == OptimizeStrategy.BATCH_PROTOBUF:
                rows = [row._pb for batch in response for row in batch.results]
            else:
                rows = [row for batch in response for row in batch.results]
            parsed_batches = []
            with futures.ThreadPoolExecutor() as executor:
                BATCH_SIZE = 10000
                batches = list(range(0, len(rows), BATCH_SIZE))
                future_to_batch = {}
                for i, batch_index in enumerate(batches):
                    try:
                        from_, to = batch_index, batches[i + 1]
                        future_to_batch[executor.submit(
                            self._parse_batch, parser,
                            itertools.islice(rows, from_,
                                             to))] = itertools.islice(
                                                 rows, from_, to)
                    except IndexError:
                        future_to_batch[executor.submit(
                            self._parse_batch, parser,
                            itertools.islice(rows, batch_index,
                                             len(rows)))] = itertools.islice(
                                                 rows, batch_index, len(rows))

                for i, future in enumerate(
                        futures.as_completed(future_to_batch), start=1):
                    batch = future_to_batch[future]
                    logger.debug(
                        "Parsed batch %d for query %s for customer_id %s", i,
                        query_specification.query_title, customer_id)
                    parsed_batch = future.result()
                    parsed_batches.append(parsed_batch)
            total_results = [
                items for batch in parsed_batches for items in batch
            ]

        else:
            logger.debug(
                "Iterating over response for query %s for customer_id %s",
                query_specification.query_title, customer_id)
            for batch in response:
                logger.debug("Parsing batch for query %s for customer_id %s",
                             query_specification.query_title, customer_id)
                if optimize_strategy == OptimizeStrategy.PROTOBUF:
                    logger.warning("Running gaarf in an optimized mode")
                    logger.warning("Optimize strategy is %s",
                                   optimize_strategy.name)
                    rows = (row._pb for row in batch.results)
                else:
                    rows = (row for row in batch.results)
                results = self._parse_batch(parser, rows)
                total_results.extend(results)
        return total_results

    def _parse_batch(
            self, parser,
            batch: Sequence["GoogleAdsRow"]) -> List[List[Tuple[Any]]]:
        """Parse reach row from batch of Ads API response.

        Args:
            parser: instance of parser class that transforms each row from
                request into desired format.
            batch: sequence of GoogleAdsRow that needs to be parsed.
        Returns:
            Parsed rows for a batch.
        """

        return [parser.parse_ads_row(row) for row in batch]


class AdsQueryExecutor:
    """Class responsible for getting data from Ads API and writing it to
        local/remote storage.

    Attributes:
        api_client: a client used for connecting to Ads API.
    """

    def __init__(self, api_client: api_clients.BaseClient):
        self.api_client = api_client

    def execute(self,
                query_text: str,
                query_name: str,
                customer_ids: Union[List[str], str],
                writer_client: writer.AbsWriter = writer.StdoutWriter(),
                args: Optional[Dict[str, Any]] = None,
                optimize_performance: str = "NONE") -> None:
        """Reads query, extract results and stores them in a specified location.

        Args:
            query_text: Text for the query.
            customer_ids: All accounts for which query will be executed.
            writer_client: Client responsible for writing data to local/remote
                location.
            args: Arguments that need to be passed to the query
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF")
        """

        query_specification = QuerySpecification(
            query_text, query_name, args,
            self.api_client.api_version).generate()
        report_fetcher = AdsReportFetcher(self.api_client, customer_ids)
        results = report_fetcher.fetch(query_specification,
                                       optimize_performance)
        logger.debug("Start writing data for query %s via %s writer",
                     query_specification.query_title, type(writer_client))
        writer_client.write(results, query_specification.query_title)
        logger.debug("Finish writing data for query %s via %s writer",
                     query_specification.query_title, type(writer_client))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple
import dataclasses
import re
from .api_clients import BaseClient
from operator import attrgetter


@dataclasses.dataclass(frozen=True)
class VirtualColumn:
    type: str
    value: str
    fields: Optional[List[str]] = None
    substitute_expression: Optional[str] = None


class VirtualColumnError(Exception):
    pass


@dataclasses.dataclass
class QueryElements:
    """Contains raw query and parsed elements.

    Attributes:
        query_title: Title of the query that needs to be parsed.
        query_text: Text of the query that needs to be parsed.
        fields: Ads API fields that need to be feched.
        column_names: friendly names for fields which are used when saving data
        customizers: Attributes of fields that need to be be extracted.
    """
    query_title: str
    query_text: str
    fields: List[str]
    column_names: List[str]
    customizers: Optional[Dict[str, Dict[str, str]]]
    virtual_columns: Optional[Dict[str, VirtualColumn]]
    resource_name: str
    is_constant_resource: bool


class QuerySpecification:

    def __init__(self,
                 text: str,
                 title: str = None,
                 args: Dict[Any, Any] = None,
                 api_version: str = "v12"):
        self.text = text
        self.title = title
        self.args = args
        self.base_client = BaseClient(api_version)

    def generate(self) -> QueryElements:
        """Reads query from a file and returns different elements of a query.

        Args:
            path: Path to a file with a query.

        Returns:
            Various elements parsed from a query (text, fields,
            column_names, etc).
        """

        query_lines = self.cleanup_query_text(self.text)
        resource_name = self.extract_resource_from_query(self.text)
        is_constant_resource = bool(resource_name.endswith("_constant"))
        query_text = " ".join(query_lines)
        if self.args:
            if (macros := self.args.get("macro")):
                query_text = query_text.format(**macros)
        fields = []
        column_names = []
        customizers = {}
        virtual_columns = {}

        query_lines = self.extract_query_lines(" ".join(query_lines))
        for line in query_lines:
            field_name = None
            customizer_type = None
            field_elements, alias, virtual_column = self.extract_fields_and_aliases(
                line)
            if field_elements:
                for field_element in field_elements:
                    field_name, customizer_type, customizer_value = \
                        self.extract_fields_and_customizers(field_element)
                    field_name = field_name.strip().replace(",", "")
                    fields.append(self.format_type_field_name(field_name))
            if not alias and not field_name:
                raise ValueError("Virtual attributes should be aliased")
            else:
                column_name = alias.strip().replace(
                    ",", "") if alias else field_name
                column_names.append(self.normalize_column_name(column_name))

            if virtual_column:
                virtual_columns[column_name] = virtual_column
            if customizer_type:
                customizers[column_name] = {
                    "type": customizer_type,
                    "value": customizer_value
                }
        query_text = self.create_query_text(fields, virtual_columns,
                                            query_text)
        return QueryElements(query_title=self.title,
                             query_text=query_text,
                             fields=fields,
                             column_names=column_names,
                             customizers=customizers,
                             virtual_columns=virtual_columns,
                             resource_name=resource_name,
                             is_constant_resource=is_constant_resource)

    def create_query_text(self, fields: List[str],
                          virtual_columns: Dict[str, VirtualColumn],
                          query_text: str) -> str:
        virtual_fields = [
            field for name, column in virtual_columns.items()
            if column.type == "expression" for field in column.fields
        ]
        if virtual_fields:
            fields = fields + virtual_fields
        query_text = f"SELECT {', '.join(fields)} {self.extract_from_statement(query_text)}"
        query_text = self._remove_traling_comma(query_text)
        query_text = self._unformat_type_field_name(query_text)
        query_text = re.sub("\s+", " ", query_text).strip()
        return query_text

    def cleanup_query_text(self, query: str) -> List[str]:
        query_lines = query.split("\n")
        result = []
        for line in query_lines:
            if re.match("^(#|--|//)", line):
                continue
            result.append(re.sub("(--|//).*$", "", line).strip())
        return result

    def extract_resource_from_query(self, query: str) -> str:
        return str(re.findall(r"FROM\s+(\w+)", query,
                              flags=re.IGNORECASE)[0]).strip()

    def extract_query_lines(self, query_text: str) -> List[str]:
        selected_fields = re.sub(r"\bSELECT\b|FROM .*",
                                 "",
                                 query_text,
                                 flags=re.IGNORECASE).split(",")
        return [field.strip() for field in selected_fields if field != " "]

    def extract_from_statement(self, query_text: str) -> str:
        return re.search(" FROM .+", query_text, re.IGNORECASE).group(0)

    def extract_fields_and_aliases(
        self, query_line: str
    ) -> Tuple[Optional[List[str]], Optional[str], Optional[VirtualColumn]]:
        fields = []
        virtual_column = None
        field_raw, *alias = re.split(" [Aa][Ss] ", query_line)
        field_raw = field_raw.replace("\s+", "").strip()
        virtual_field, _, _ = self.extract_fields_and_customizers(field_raw)
        try:
            _ = attrgetter(virtual_field)(self.base_client.google_ads_row)
        except AttributeError:
            operators = ("/", "\*", "\+", "-")
            if len(expressions := re.split("|".join(operators),
                                           field_raw)) > 1:
                virtual_column_fields = []
                substitute_expression = virtual_field
                for element in expressions:
                    element = element.strip()
                    try:
                        _ = attrgetter(element)(
                            self.base_client.google_ads_row)
                        virtual_column_fields.append(element)
                        substitute_expression = substitute_expression.replace(
                            element, f"{{{element}}}")
                    except AttributeError:
                        pass
                virtual_column = VirtualColumn(
                    type="expression",
                    value=virtual_field,
                    fields=virtual_column_fields,
                    substitute_expression=substitute_expression.replace(".", "_"))
            else:
                virtual_column = VirtualColumn(type="built-in",
                                                     value=virtual_field)
        if not virtual_column and field_raw:
            fields = [field_raw]
        else:
            fields = None
        return fields, alias[0] if alias else None, virtual_column

    def extract_fields_and_customizers(self, line_elements: str):
        resources = self.extract_resource_element(line_elements)
        pointers = self.extract_pointer(line_elements)
        nested_fields = self.extract_nested_resource(line_elements)
        if len(resources) > 1:
            field_name, resource_index = resources
            return field_name, "resource_index", int(resource_index)
        if len(pointers) > 1:
            field_name, pointer = pointers
            return field_name, "pointer", pointer
        if len(nested_fields) > 1:
            field_name, nested_field = nested_fields
            return field_name, "nested_field", nested_field
        return line_elements, None, None

    def extract_resource_element(self, line_elements: str) -> List[str]:
        return re.split("~", line_elements)

    def extract_pointer(self, line_elements: str) -> List[str]:
        return re.split("->", line_elements)

    def extract_nested_resource(self, line_elements: str) -> List[str]:
        return re.split(":", line_elements)

    def format_type_field_name(self, field_name: str) -> str:
        return re.sub(r"\.type", ".type_", field_name)

    def normalize_column_name(self, column_name: str) -> str:
        return re.sub(r"\.", "_", column_name)

    def _remove_traling_comma(self, query: str) -> str:
        return re.sub(r",\s+from", " FROM", query, re.IGNORECASE)

    def _unformat_type_field_name(self, query: str) -> str:
        return re.sub(r"\.type_", ".type", query)



import pytest
import gaarf.query_editor as query_editor


@pytest.fixture
def query_specification():
    query = """
-- Comment
# Comment
// Comment

SELECT
    '20230101' AS date,
    metrics.clicks / metrics.impressions AS ctr,
    customer.id, --customer_id
    campaign.bidding_strategy_type AS campaign_type, campaign.id:nested AS campaign,
    ad_group.id~1 AS ad_group,
    ad_group_ad.ad.id->asset AS ad,
    metrics.cost_micros * 1e6 AS cost,
    campaign.selective_optimization AS selective_optimization,
from ad_group_ad
"""
    return query_editor.QuerySpecification(title="sample_query",
                                           text=query,
                                           args=None)


@pytest.fixture
def sample_query(query_specification):
    return query_specification.generate()


@pytest.fixture
def incorrect_query_specification():
    query = """
SELECT
    "${custom_field}",
    ad_group.id AS ad_group_id
FROM ad_group_id
"""
    return query_editor.QuerySpecification(title="sample_query",
                                           text=query,
                                           args=None)


def test_correct_title(sample_query):
    assert sample_query.query_title == "sample_query"


def test_extract_correct_fields(sample_query):
    assert sample_query.fields == [
        "customer.id", "campaign.bidding_strategy_type", "campaign.id",
        "ad_group.id", "ad_group_ad.ad.id", "campaign.selective_optimization"
    ]


def test_extract_correct_aliases(sample_query):
    assert sample_query.column_names == [
        "date", "ctr", "customer_id", "campaign_type", "campaign", "ad_group",
        "ad", "cost", "selective_optimization"
    ]


def test_extract_correct_text(sample_query):
    assert sample_query.query_text.lower(
    ) == "select customer.id, campaign.bidding_strategy_type, campaign.id, ad_group.id, ad_group_ad.ad.id, campaign.selective_optimization, metrics.clicks, metrics.impressions, metrics.cost_micros from ad_group_ad"


def test_extract_custom_callers(sample_query):
    assert sample_query.customizers == {
        "campaign": {
            "type": "nested_field",
            "value": "nested"
        },
        "ad_group": {
            "type": "resource_index",
            "value": 1
        },
        "ad": {
            "type": "pointer",
            "value": "asset"
        }
    }


def test_extract_query_lines(query_specification, sample_query):
    extracted_lines = query_specification.extract_query_lines(
        sample_query.query_text)
    assert extracted_lines == [
        "customer.id", "campaign.bidding_strategy_type", "campaign.id",
        "ad_group.id", "ad_group_ad.ad.id", "campaign.selective_optimization",
        "metrics.clicks", "metrics.impressions", "metrics.cost_micros"
    ]


def test_extract_correct_resource(sample_query):
    assert sample_query.resource_name == "ad_group_ad"


def test_is_constant_resource(sample_query):
    assert sample_query.is_constant_resource == False


def test_has_virtual_columns(sample_query):
    assert sample_query.virtual_columns == {
        "date":
        query_editor.VirtualColumn(type="built-in", value="'20230101'"),
        "ctr":
        query_editor.VirtualColumn(
            type="expression",
            value="metrics.clicks / metrics.impressions",
            fields=["metrics.clicks", "metrics.impressions"],
            substitute_expression="{metrics_clicks} / {metrics_impressions}"),
        "cost":
        query_editor.VirtualColumn(
            type="expression",
            value="metrics.cost_micros * 1e6",
            fields=["metrics.cost_micros"],
            substitute_expression="{metrics_cost_micros} * 1e6")
    }


def test_incorrect_specification_raises_value_error(
        incorrect_query_specification):
    with pytest.raises(ValueError):
        incorrect_query_specification.generate()



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Sequence, Tuple, Union
import random
import string
from faker import Faker

from gaarf.report import GaarfReport
from gaarf.query_executor import AdsReportFetcher
from gaarf.api_clients import BaseClient
from gaarf.query_editor import QuerySpecification


@dataclass(frozen=True)
class FakeField:
    type: Any
    name: str


@dataclass
class SimulatorSpecification:
    api_version: str = "v12"
    n_rows: int = 1000
    days_ago: str = "-7d"
    string_length: int = 3
    ignored_enums: Tuple[str, ...] = ("REMOVED", "UNKNOWN", "UNSPECIFIED")
    allowed_enums: Optional[Dict[str, Sequence[str]]] = None
    replacements: Optional[Dict[str, Any]] = None


def simulate_data(
    query: str,
    simulator_specification: SimulatorSpecification,
    simulation_helper: Callable = Faker()
) -> GaarfReport:
    query_specification = QuerySpecification(query).generate()
    client = BaseClient(simulator_specification.api_version)
    report_fetcher = AdsReportFetcher(client, [])
    other_types = client.infer_types(query_specification.fields)
    v = report_fetcher.fetch(query_specification)
    try:
        iter(v.results[0])
        results = v.results[0]
    except TypeError:
        results = [v.results[0]]
    if not results:
        results = [results]
    types = [
        FakeField(type(field_value), field_name)
        for field_value, field_name in zip(results, query_specification.fields)
    ]
    values = []
    for _ in range(simulator_specification.n_rows):
        row = []
        for i, field in enumerate(types):
            if len(enums := other_types[i].values) > 1:
                if allowed_enums := simulator_specification.allowed_enums:
                    if selected_enums := allowed_enums.get(field.name):
                        value = random.choice(selected_enums)
                    else:
                        value = random.choice(list(enums))
                else:
                    value = random.choice([
                        value for value in other_types[i].values
                        if value not in simulator_specification.ignored_enums
                    ])
            elif (simulator_specification.replacements
                  and field.name in simulator_specification.replacements):
                value = random.choice(
                    simulator_specification.replacements[field.name])
            else:
                value = generate_random_value(
                    field, simulator_specification, simulation_helper,
                    query_specification.column_names[i])
            row.append(value)
        values.append(row)
    return GaarfReport(values, query_specification.column_names)


def generate_random_value(field: FakeField,
                          simulator_specification: SimulatorSpecification,
                          simulation_helper: Callable,
                          column_name: str) -> Union[bool, str, float]:
    if field.type == str:
        if "date" in field.name:
            return simulation_helper.date_between(
                simulator_specification.days_ago).strftime("%Y-%m-%d")
        if "url" in field.name:
            return "example.com"
        if "video.id" in field.name or "video_id" in field.name:
            return "4WXs3sKu41I"
        if "id" in column_name:
            return str(random.randint(1000000, 1000010))
        return generate_random_string(simulator_specification.string_length)
    if field.type == int:
        if ".id" in field.name:
            return random.randint(1000000, 1000010)
        if "micros" in field.name:
            return int(random.randint(0, 1000) * 1e6)
        return random.randint(0, 1000)
    if field.type == float:
        return float(random.randint(0, 1000))
    if field.type == bool:
        return bool(random.getrandbits(1))
    return ""


def generate_random_string(length: int) -> str:
    characters = string.ascii_lowercase
    return ''.join(random.choice(characters) for _ in range(length))
